/** Record<string, DataType> alias */
type DataRecord = {
    [P in string]: DataType;
};
/** Defining posible data type rather than using unknown */
type DataType = DataRecord | DataRecord[] | DataType[] | string | number | boolean | null | undefined;
type OmitUtilityProps<K> = K extends string ? K extends `${'$'}${infer _X}` | 'toJSON' ? never : K : K;

type parseOptions<T> = {
    default?: T;
    parseHandler?: (value: DataType) => T;
};

/**
 * Inferer: safely infering data for managing and transforming data
 */
declare class Inferer {
    _data: DataType;
    constructor(data: DataType);
    /**
     * create error object
     *
     * @private
     * @param {...string[]} tags
     * @return {Error}
     * @memberof Inferer
     */
    private _createError;
    /**
     * fallback handler for each infer type function
     *
     * @private
     * @template T
     * @param {string} path
     * @param {string} type
     * @param {DataType} value
     * @param {parseOptions<T>} options
     * @return {T}
     * @memberof Inferer
     */
    private _fallbackHandler;
    /**
     * Get raw value based on object path
     */
    get(path?: string): DataType;
    /**
     * check given path if value exist or not.
     *
     * @param {string} path
     * @param {(data: DataType) => boolean} [validator=(v) => v !== undefined] default to `(v) => v !== undefined`
     * @return {boolean}
     */
    has(path: string, validator?: (data: DataType) => boolean): boolean;
    /**
     * Infer object path in runtime either value is string or throw an error
     *
     * @param {string} [path='']
     * @param {parseOptions<string>} [options={}]
     * @return {string}
     * @memberof Inferer
     */
    string(path?: string, options?: parseOptions<string>): string;
    /**
     * Infer object path in runtime either value is number (int|float) or throw an error
     *
     * @param {string} [path='']
     * @param {(parseOptions<number> & { float?: boolean })} [options={}]
     * @return {number}
     * @memberof Inferer
     */
    number(path?: string, options?: parseOptions<number> & {
        float?: boolean;
    }): number;
    /**
     * Infer object path in runtime either value is boolean or throw an error
     *
     * @param {string} [path='']
     * @param {parseOptions<boolean>} [options={}]
     * @return {boolean}
     * @memberof Inferer
     */
    boolean(path?: string, options?: parseOptions<boolean>): boolean;
    /**
     * Infer object path in runtime either value is string | number or throw an error
     *
     * @param {string} [path='']
     * @param {parseOptions<string | number>} [options={}]
     * @return {string | number}
     * @memberof Inferer
     */
    stringOrNumber(path?: string, options?: parseOptions<string | number>): string | number;
    /** alias for stringOrNumber */
    snum(path?: string, options?: parseOptions<string | number>): string | number;
    /**
     * Infer object path in runtime is array then return as InfererArray object.
     *
     * @param {boolean} [defaultEmpty=false] when true return will fallback to empty array. On false will throw error. default: false.
     */
    array(path?: string, defaultEmpty?: boolean): InfererArray;
    /**
     * Infer object path when path has value it will parse or transform with given function or throw an error when value is null or undefined.
     */
    parse<T = DataType>(path: string, fn: (data: unknown) => T, options?: parseOptions<T>): T;
    /**
     * Infer object path and fallback to given preset value and type if type not match
     */
    maybe<T>(path: string, preset: T): {
        /** check if given path is a string or return preset value */
        string(options?: parseOptions<string>): string | T;
        /** check if given path is a number or return preset value */
        number(options?: parseOptions<number> & {
            float?: boolean;
        }): number | T;
        /** check if given path is a boolean or return preset value */
        boolean(options?: parseOptions<boolean>): boolean | T;
        /** check if given path is a string, number or return preset value */
        stringOrNumber(options?: parseOptions<string | number>): string | number | T;
        /** check if given path has value then parse using given function or return preset value */
        parse<V>(fn: (data: unknown) => V, options?: parseOptions<V>): T | V;
        /** check if given path is array then return InfererArray or return preset value */
        array(): InfererArray | T;
    };
    /**
     * alias or shortcut for maybe with preset value null
     */
    nullable(path?: string): {
        /** check if given path is a string or return preset value */
        string(options?: parseOptions<string>): string | null;
        /** check if given path is a number or return preset value */
        number(options?: parseOptions<number> & {
            float?: boolean | undefined;
        }): number | null;
        /** check if given path is a boolean or return preset value */
        boolean(options?: parseOptions<boolean>): boolean | null;
        /** check if given path is a string, number or return preset value */
        stringOrNumber(options?: parseOptions<string | number>): string | number | null;
        /** check if given path has value then parse using given function or return preset value */
        parse<V>(fn: (data: unknown) => V, options?: parseOptions<V>): V | null;
        /** check if given path is array then return InfererArray or return preset value */
        array(): InfererArray | null;
    };
}
/**
 * Extended array object with typesafe infering and transform capability
 *
 * @export
 * @class InfererArray
 * @extends {Array}
 */
declare class InfererArray extends Array {
    constructor(collection: DataType[]);
    /**
     * loop each item and parse or transform item using given function.
     *
     * @param {(row: Inferer, index: number, arr: InfererArray) => T} fn
     * @return {T[]} new transformed array
     */
    inferEach<T = DataType>(fn: (row: Inferer, index: number, arr: InfererArray) => T): T[];
    /**
     * extended Array.at that return inferer object.
     *
     * @param {number} index The zero-based index of the desired code unit. A negative index will count back from the last item.
     */
    inferAt(index: number): Inferer;
}

/**
 * Typesafe infer and transform data.
 */
declare function infer(data: unknown): Inferer;

declare class SchemaNode {
    /** Generated string scope */
    $scope: string;
    /** Description for the string scope */
    $description: string;
    /**
     * overriding json stringify
     *
     * @see https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/JSON/stringify#tojson_behavior
     */
    toJSON(): unknown;
    constructor(scope: string, description: string);
}
declare class PermitNode extends SchemaNode {
    /** access control value for scope */
    private $value;
    /** hook function for subscribing to scope changes */
    private $subscriber;
    constructor(schema: SchemaNode, value?: boolean);
    /** helper for broadcasting changes to subscriber */
    private _broadcast;
    toJSON(): {
        scope: string;
        description: string;
        value: boolean;
    };
    /** readonly boolean value of true on disabled */
    get $disable(): boolean;
    /** boolean value of the scope. will broadcast changes when assigned new value to event listener. */
    get $able(): boolean;
    /** boolean value of the scope. will broadcast changes when assigned new value to event listener. */
    set $able(value: boolean);
    /** add listener function to scope that called on value changes */
    $subscribe(fn: (active: boolean, scope: SchemaScope) => void): () => boolean;
    /** utility for toggling value. will broadcast changes every call to event listener. */
    $toggle(): void;
}

declare class SchemaBranch<T extends GeneratedSchema> {
    readonly $schema: T;
    private _meta;
    constructor(schema: T, scope?: string, spacer?: string);
    toJSON(): unknown;
    get $meta(): {
        scope: string;
        spacer: string;
    };
    get $scope(): string;
    $each(nodeHook: (node: SchemaNode, key: string) => void, branchHook?: (branch: SchemaBranch<any>, key: string) => void): void;
    $refactor(prefix?: string, spacer?: string): void;
    $export(): {
        schema: SchemaExport<T>;
        meta: {
            scope: string;
            spacer: string;
        };
    };
    $model(): RevertSchema<T>;
    $permit(scopes: (string | SchemaNode)[]): Permit<T>;
}

type SchemaScope = {
    $scope: string;
    $description: string;
};
/** Tree object for defining access control schema */
interface SchemaTree<T = string> extends Record<string, SchemaTree<T> | T> {
}
/** Generated tree schema type */
type GeneratedSchema<T extends SchemaTree = SchemaTree> = T extends SchemaTree ? {
    [P in keyof T]: T[P] extends SchemaTree ? Schema<GeneratedSchema<T[P]>> : SchemaNode;
} : never;
type RevertSchema<T> = T extends Record<string, any> ? {
    [K in OmitUtilityProps<keyof T>]: T[K] extends SchemaNode ? string : RevertSchema<T[K]>;
} : never;
type SchemaExport<T> = T extends Record<string, any> ? {
    [K in OmitUtilityProps<keyof T>]: T[K] extends SchemaNode ? {
        $scope: string;
        $description: string;
    } : SchemaExport<T[K]>;
} : never;
/** Access control schema tree */
type Schema<T extends GeneratedSchema> = T & SchemaBranch<T>;
type PermitNodes<T = Record<string, any>> = T extends Record<string, any> ? {
    [K in OmitUtilityProps<keyof T>]: T[K] extends SchemaNode ? PermitNode : T[K] extends {
        $scope: string;
        $description: string;
    } ? PermitNode : PermitNodes<T[K]> & PermitBranch<T[K]>;
} : never;
type PermitSchema<T extends Record<string, any>> = PermitNodes<T> & PermitBranch<T>;
/**
 * Infering literal string to scopes
 *
 * @template T
 */
type InferLiteralString<T, TD extends string> = T extends infer K ? K extends string ? K extends '' ? `${TD}` : `${K}` : `${TD}` : `${TD}`;
type ScopePrefixer<T, S> = T extends infer K ? K extends string ? K extends '' ? '' : `${K}${InferLiteralString<S, '.'>}` : never : never;
type UnionScopes<T, Prefix extends string = '', Spacer extends string = '.'> = T extends Record<string, unknown> ? {
    [K in keyof T]: K extends string ? T[K] extends Record<string, unknown> ? UnionScopes<T[K], `${ScopePrefixer<Prefix, Spacer>}${K}`, InferLiteralString<Spacer, '.'>> : `${ScopePrefixer<Prefix, Spacer>}${K}` : never;
}[keyof T] : Prefix;
type SchemaOptions = {
    /** add prefix to scope. default to '' (none) */
    prefix: string;
    /** string to use for spacer for scope. default to '.' */
    spacer: string;
};

declare class Permit<T extends SchemaTree<any>> {
    /** Access Schema */
    readonly on: PermitSchema<T>;
    private _nodes;
    private schema;
    private scopes;
    constructor(schema: T, scopes?: string[]);
    static getNodeScope(scope: string | SchemaNode | PermitNode): string;
    /** get nodes that has active access or granted */
    get granted(): PermitNode[];
    /** get nodes that is inactive or denied access */
    get denied(): PermitNode[];
    /** get list of permit node */
    get nodes(): PermitNode[];
    /** reset scope to default state */
    reset(): void;
    /** grant access to scope */
    grant(...scopes: (string | SchemaNode)[]): void;
    /** deny access to scope */
    deny(...scopes: (string | SchemaNode | PermitNode)[]): void;
    ableTo(scope: string | SchemaNode | PermitNode): boolean;
    get(scope: string | SchemaNode): PermitNode | undefined;
    has(scope: string | SchemaNode | PermitNode): boolean;
    child(scope: string): PermitNode[];
    /** loop through each node */
    $each(fn: (node: PermitNode) => void): void;
}
declare class PermitBranch<T extends Record<string, any>> {
    private $tree;
    constructor(nodes: PermitNodes<T>);
    $each(fn: (node: PermitNode) => void): void;
    get $able(): boolean;
    get $doable(): boolean;
    get $disable(): boolean;
    get $nodes(): PermitNode[];
    get $granted(): PermitNode[];
    get $denied(): PermitNode[];
}

/**
 * create access control schema from tree model. Return generated access control.
 *
 * ```js
 *
 * 	const schema = createSchema({
 * 		tweet: {
 * 			create: "create a blog post",
 * 			edit: "edit a blog post",
 * 			delete: "delete a blog post"
 * 		}
 * 	}, { prefix: 'com.tweeter', spacer: ':' })
 *
 * 	console.log(schema.tweet.create.$scope)
 * 	// output: 'com.tweeter:tweet:create'
 *
 *
 * ```
 */
declare function createSchema<T extends SchemaTree<string>>(model: T, options?: Partial<SchemaOptions>): Schema<GeneratedSchema<T>>;
/**
 * create permit object from exported schema.
 * usefull for ssr application when only need to create a permit on requested page.
 *
 * ```ts
 *
 * // on server side, send json data of schema access
 * const blogPageSchema = schema.blog.export().schema
 * const json = JSON.stringify(blogPageSchema)
 * type BlogPageSchema = typeof blogPageSchema
 *
 * // on client, parse json data
 * const schema = JSON.parse(json)
 * const permit = createPermit<BlogPageSchema>(schema, [])
 *
 *
 * ```
 */
declare function createPermit<T extends SchemaTree<SchemaScope>>(schema: T, scopes: string[]): Permit<T>;

/** General context type must object with string key. */
type Ctx = Record<string, any>;
/** Context with previous value inside, used in pipe function. */
type ContextPipe<T extends Ctx, V> = T & {
    value: V;
};
/** Pipe context injector. Used under procedure or pipe wrap function for resolving end value. */
type PipeInjector<T extends Ctx, V> = {
    context: T;
    value: V;
};
/**
 * garda procedure types.
 *
 * @template O Output types.
 * @template I Array type of input arguments.
 * @template T Context types.
 */
type Procedure<O, I extends any[], T extends Ctx> = ((...input: I) => O) & {
    /**
     * garda pipe parse utility function to parse output into new procedure with new value.
     *
     * ---
     *
     * ```js
     *
     * // Usage
     * const getContextCount = context
     *  .procedure((ctx) => ctx.getCount())
     *  .parse((data) => typeof data === 'number' ? data : 0)
     *
     * // type count is number
     * const count = getContextCount()
     *
     * // Usage with zod
     * const getZodContextCount = context
     *   .procedure((ctx) => ctx.getCount())
     *   .parse(z.number().parse) // or z.number().safeParse
     *
     * // type value is number
     * const value = getZodContextCount()
     *
     * ```
     *
     *  */
    parse: <V>(parser: (data: any) => V) => Procedure<V, I, T>;
    /**
     * garda pipe utility function to parse output into new procedure with new value.
     * Equal to parse but for handling promises output.
     *
     * ---
     *
     * ```js
     *
     * // Usage
     * const getUser = context
     *  .procedure(async (ctx, userid) => await ctx.fetchUserData(userid))
     *  .parseAsync((data) => {
     *    if (!data) return undefined
     *    // some validation or transformation here
     *    return data
     *  })
     *
     * const user = await getUser("some-id")
     *
     * // Usage with zod
     * const getUserData = context
     *   .procedure(async (ctx, userid) => await ctx.fetchUserData(userid))
     *   .parse(userDataSchema.parse) // or userDataSchema.safeParse | parseAsync | spa
     *
     * const userdata = await getUserData("some-id")
     *
     * ```
     *
     *  */
    parseAsync: <V>(parser: (data: any) => V) => Procedure<Promise<Awaited<V>>, I, T>;
    /**
     * garda pipe utility function for infering output and transform to new output type.
     *
     * ---
     *
     * ```ts
     *
     * // Usage
     * const getUserState = context
     *  .procedure((ctx, key: string) => ctx.getLocalStorageState(key))
     *  .infer(inferer => ({
     *    status: inferer.string('user.status.label'),
     *    hasWife: inferer.string('user.partner.gender' === 'female',
     *    hasChild: inferer.number('user.children.count') > 0
     *  }))
     *
     * const state = getUserState('some-key')
     *
     * ```
     */
    infer: <V>(inferFn: (inferer: Inferer) => V) => Procedure<V, I, T>;
    /**
     * garda pipe utility function for infering output and transform to new output type.
     * Equal to infer but for handling promises output.
     *
     * ---
     *
     * ```ts
     *
     * // Usage
     * const getUserName = context
     *  .procedure(async (ctx, id: string) => await ctx.fetchUserData(id))
     *  .inferAsync(inferer =>
     *    inferer.string('data.fullName') + ' ' + inferer.string('data.lastName')
     *  )
     *
     * const name = await getUserName('some-user-id')
     *
     * ```
     */
    inferAsync: <V>(inferFn: (inferer: Inferer) => V) => Procedure<Promise<Awaited<V>>, I, T>;
    /**
     * garda utility function for transforming and processing output from procedure.
     *
     * ---
     *
     * ```js
     *
     * // Usage
     * const getCount = context
     *  .procedure((ctx) => ctx.count)
     *
     * const doubleCount = getCount.pipe(({ value }) => value * 2)
     * const isCountEven = getCount.pipe(({ value }) => value % 2)
     *
     * const count = getCount() // return 1
     * const doubled = doubleCount() // return 2
     * const isEven = isCountEven() // return true
     *
     * ```
     */
    pipe: <V>(pipeFn: (context: ContextPipe<T, O>, ...input: I) => V) => Procedure<V, I, T>;
    /**
     * garda utility for injecting updated context.
     *
     * useful when context is globally stored and updated,
     * but procedure need to be isolated.
     *
     * ---
     *
     * ```js
     *
     * // Example
     *
     * // isolated get data procedure with fetch
     * const getData = context
     *  .procedure((ctx) => fetch('/some-api', {
     *    headers: { "Authorization": `Bearer ${ctx.token}` }
     *  }))
     *
     * // get context and set token from cookie
     * const ctx = context()
     * ctx.token = cookie.get('Access-Token')
     * getData.inject(ctx) // inject updated context
     * const data = await getData()
     *
     * // or using context update utility
     * context.update((ctx) => {
     *  ctx.token = cookie.get('Access-Token)
     *  return ctx
     * })
     *
     * getData.inject(context)
     *
     * // or even better, directly update procedure context
     * getData.inject(ctx => {
     *  ctx.token = cookie.get('Access-Token)
     *  return ctx
     * })
     * ```
     */
    inject: (context: T | Context<T> | ((context: T) => void)) => Procedure<O, I, T>;
    /**
     * garda utility for reseting context that has been injected to current procedure.
     */
    resetContext: () => Procedure<O, I, T>;
};
type ProcedureOptions = {
    onAfterCalled: (value: any) => void;
    onBeforeCalled: (input: any[]) => void;
};
type Context<T extends Record<string, any> = any> = (() => T) & {
    /**
     * context function, alias for context()
     *
     * > ! notice. Different treatment on immidiate false/true.
     * >
     * > on immidiate option false (default), `context()` and `context.get()`
     * > will create a fresh context everytime. But when immidiate is set to true,
     * > context is already initiated. So it's not going to be isolated context.
     * */
    get: () => T;
    /**
     * context utility for set context value
     *
     * > ! notice. Different tretment on immidiate false/true.
     * >
     * > on immidiate option true, when set is called. context isolated feature will be
     * > terminated because context is set to object that already initiated. And every update
     * > on context and it's object that context set to will be immidiately applied.
     * >
     * > Isolated feature of the context can be set again by calling `context.reset()`,
     * > and context will be set to initial context value.
     */
    set: (context: T) => void;
    /**
     * context utility for updating context value
     *
     * > ! notice. Different tretment on immidiate false/true.
     * >
     * > on immidiate option false (default). Given update will be treated as a hook function
     * > that only will be called when procedure is called as needed because of the isolated treatment.
     * >
     * > But when isolated feature is terminated by setting context through `context.set()`,
     * > update will be applied immidiately.
     */
    update: (fn: (context: T) => void) => void;
    /**
     * reset context to initial context initiation function.
     * this will remove all previous update to the context.
     * on immidiate set to true, context that is reseted will not point to previous initiated context.
     *
     * ```js
     * // Example
     * const context = createContext(() => {
     *  let count = 0
     *  return { setCount: (num) => count = num, count: () => count }
     * }, { immidiate: true })
     *
     * const ctx = context.get()
     *
     * ctx.setCount(10)
     * context().count() // this will return 10 too.
     *
     * context.update(c => c.setCount(20))
     * ctx.count() // this will return 20
     *
     * // on reset context
     * context.reset()
     * context().count() // will return 0
     * ctx.count() // will return 20
     *
     * ```
     */
    reset: () => void;
    /** create procedure from context */
    procedure: <I extends any[], V>(fn: (context: T, ...input: I) => V) => Procedure<V, I, T>;
};
type ContextOptions = {
    /**
     * Option for immidiately run context.
     * useful when context should be shared between procedure.
     *
     * When false, context will be isolated from each procedure and run when only needed.
     *
     * Default to false.
     */
    immidiate: boolean;
} & ProcedureOptions;
type InferContext<C extends Context> = C extends Context<infer T> ? T : never;

/**
 * create garda procedure
 */
declare function createProcedure<O, I extends any[], T extends Record<string, any>>(context: (() => T) | Context<T>, fn: (context: T, ...input: I) => O, options?: Partial<ProcedureOptions>): Procedure<O, I, T>;
/** Create garda context */
declare function createContext<T extends Record<string, any>>(context: () => T, options?: Partial<ContextOptions>): Context<T>;

export { Context, ContextOptions, ContextPipe, InferContext, Inferer, InfererArray, Permit, PermitBranch, PermitNode, PipeInjector, Procedure, ProcedureOptions, Schema, SchemaBranch, SchemaNode, SchemaOptions, SchemaTree, UnionScopes, createContext, createPermit, createProcedure, createSchema, infer, parseOptions };
