/* eslint-disable @typescript-eslint/no-unused-vars */
/* eslint-disable @typescript-eslint/no-empty-function */
/* eslint-disable @typescript-eslint/no-explicit-any */

import type { DataType } from './../types';
import { type Inferer, infer } from './../infer/index';
import type {
  ProcedureOptions,
  PipeInjector,
  Procedure,
  ContextOptions,
  ContextPipe,
  Context,
} from './types';

// export all types
export * from './types';

/**
 * create garda procedure
 */
export function createProcedure<
  O,
  I extends any[],
  T extends Record<string, any>
>(
  context: (() => T) | Context<T>,
  fn: (context: T, ...input: I) => O,
  options: Partial<ProcedureOptions> = {}
) {
  // keep original context for reset
  /** actively used context */
  let _context = context;

  // recursive pipe generator
  const generatePipe = <Vprev, Vnext>(
    proc: (...input: I) => PipeInjector<T, Vprev>,
    pipe: (context: ContextPipe<T, Vprev>, ...input: I) => Vnext
  ): Procedure<Vnext, I, T> => {
    // pipe injector for context and previous value
    const injector = (...input: I) => {
      const pipeReturn = proc(...input);
      const value = pipe(
        { ...pipeReturn.context, value: pipeReturn.value },
        ...input
      );
      return { context: pipeReturn.context, value };
    };

    const pipes = Object.assign(
      (...input: I) => {
        const { context, value } = proc(...input);
        const finalValue = pipe({ ...context, value }, ...input);

        if (typeof options.onAfterCalled === 'function') {
          options.onAfterCalled(finalValue);
        }

        return finalValue;
      },
      {
        pipe: <V>(
          pipeFn: (context: ContextPipe<T, Vnext>, ...input: I) => V
        ) => {
          return generatePipe<Vnext, V>(injector, pipeFn) as Procedure<V, I, T>;
        },

        infer: <V>(inferFn: (inferer: Inferer) => V) => {
          // eslint-disable-next-line @typescript-eslint/no-unused-vars
          return generatePipe<Vnext, V>(injector, ({ value }, ..._: I) => {
            return inferFn(infer(value as DataType));
          }) as Procedure<V, I, T>;
        },

        inferAsync: <V>(inferFn: (inferer: Inferer) => Promise<V>) => {
          // eslint-disable-next-line @typescript-eslint/no-unused-vars
          return generatePipe<Vnext, Promise<V>>(
            injector,
            async ({ value }, ..._: I) => {
              const awaitedValue = await Promise.resolve(value);
              return inferFn(infer(awaitedValue as DataType));
            }
          ) as Procedure<Promise<V>, I, T>;
        },

        parse: <V>(parser: (data: any) => V) => {
          // eslint-disable-next-line @typescript-eslint/no-unused-vars
          return generatePipe<Vnext, V>(injector, ({ value }, ..._: I) => {
            return parser(value);
          }) as Procedure<V, I, T>;
        },

        parseAsync: <V>(parser: (data: any) => Promise<V>) => {
          // eslint-disable-next-line @typescript-eslint/no-unused-vars
          return generatePipe<Vnext, Promise<V>>(
            injector,
            async ({ value }, ..._: I) => {
              const awaitedValue = await Promise.resolve(value);
              return parser(awaitedValue);
            }
          ) as Procedure<Promise<V>, I, T>;
        },

        inject: (ctx: T | Context<T> | ((ctx: T) => void)) => {
          // set context
          if (typeof ctx === 'function') {
            if (Object.hasOwn(ctx, 'get') && Object.hasOwn(ctx, 'procedure')) {
              _context = ctx as Context<T>;
            } else {
              const _ctx = _context();
              ctx(_ctx);

              _context = () => _ctx;
            }
          } else _context = () => ctx;

          // passthrough previous pipe
          return generatePipe<Vprev, Vnext>(proc, pipe);
        },

        resetContext: () => {
          _context = context;
          return generatePipe<Vprev, Vnext>(proc, pipe);
        },
      }
    ) as Procedure<Vnext, I, T>;

    // return as new procedure
    return pipes;
  };

  // create initial injector
  const injector = (...input: I) => {
    const ctx = _context();
    if (typeof options.onBeforeCalled === 'function') {
      options.onBeforeCalled(input);
    }

    return {
      context: ctx,
      value: fn(ctx, ...input),
    };
  };

  // prepare caller
  const procedure = Object.assign(
    (...input: I) => {
      const ctx = _context();
      if (typeof options.onBeforeCalled === 'function') {
        options.onBeforeCalled(input);
      }

      const value = fn(ctx, ...input);

      if (typeof options.onAfterCalled === 'function') {
        options.onAfterCalled(value);
      }

      return value;
    },
    {
      // create first pipe function
      pipe: <V>(pipeFn: (context: ContextPipe<T, O>, ...input: I) => V) => {
        return generatePipe<O, V>(injector, pipeFn) as Procedure<V, I, T>;
      },

      // create first infer function
      infer: <V>(inferFn: (inferer: Inferer) => V) => {
        // eslint-disable-next-line @typescript-eslint/no-unused-vars
        return generatePipe<O, V>(injector, ({ value }, ..._: I) => {
          return inferFn(infer(value as DataType));
        }) as Procedure<V, I, T>;
      },

      inferAsync: <V>(inferFn: (inferer: Inferer) => Promise<V>) => {
        // eslint-disable-next-line @typescript-eslint/no-unused-vars
        return generatePipe<O, Promise<V>>(
          injector,
          async ({ value }, ..._: I) => {
            const awaitedValue = await Promise.resolve(value);
            return inferFn(infer(awaitedValue as DataType));
          }
        ) as Procedure<Promise<V>, I, T>;
      },

      parse: <V>(parser: (data: any) => V) => {
        // eslint-disable-next-line @typescript-eslint/no-unused-vars
        return generatePipe<O, V>(injector, ({ value }, ..._: I) => {
          return parser(value);
        }) as Procedure<V, I, T>;
      },

      parseAsync: <V>(parser: (data: any) => Promise<V>) => {
        // eslint-disable-next-line @typescript-eslint/no-unused-vars
        return generatePipe<O, Promise<V>>(
          injector,
          async ({ value }, ..._: I) => {
            const awaitedValue = await Promise.resolve(value);
            return parser(awaitedValue);
          }
        ) as Procedure<Promise<V>, I, T>;
      },

      inject: (ctx: T | Context<T> | ((ctx: T) => void)) => {
        // set context
        if (typeof ctx === 'function') {
          if (Object.hasOwn(ctx, 'get') && Object.hasOwn(ctx, 'procedure')) {
            _context = ctx as Context<T>;
          } else {
            const _ctx = _context();
            ctx(_ctx);

            _context = () => _ctx;
          }
        } else _context = () => ctx;

        // create new procedure
        return createProcedure<O, I, T>(_context, fn, options);
      },

      resetContext: () => {
        _context = context;
        return createProcedure<O, I, T>(_context, fn, options);
      },
    }
  ) as Procedure<O, I, T>;

  return procedure;
}

/** Create garda context */
export function createContext<T extends Record<string, any>>(
  context: () => T,
  options: Partial<ContextOptions> = {}
) {
  const { immidiate, ...procOptions } = options;

  if (immidiate === true) {
    let $ctx = context();

    const sharedContext = Object.assign(() => $ctx, {
      get: () => $ctx,
      set(ctx: T) {
        $ctx = ctx;
      },
      update(fn: (ctx: T) => void) {
        fn($ctx);
      },
      reset() {
        $ctx = context();
      },
      procedure<V, I extends any[]>(fn: (context: T, ...input: I) => V) {
        return createProcedure<V, I, T>(() => $ctx, fn, procOptions);
      },
    });

    return sharedContext as Context<T>;
  } else {
    let $ctx: T | null = null;
    const contextUpdate = new Set<(ctx: T) => void>();
    const _context = () => {
      if ($ctx !== null) return $ctx;

      const _ctx = context();
      for (const fn of contextUpdate.values()) {
        fn(_ctx);
      }

      return _ctx;
    };

    const isolatedContext = Object.assign(() => _context(), {
      get: () => _context(),
      set(ctx: T) {
        $ctx = ctx;
        contextUpdate.clear();
      },
      update(fn: (ctx: T) => void) {
        if ($ctx !== null) fn($ctx);
        else contextUpdate.add(fn);
      },
      reset() {
        $ctx = null;
        contextUpdate.clear();
      },
      procedure<V, I extends any[]>(fn: (context: T, ...input: I) => V) {
        return createProcedure<V, I, T>(() => _context(), fn, procOptions);
      },
    });

    return isolatedContext as Context<T>;
  }
}
