/* eslint-disable @typescript-eslint/no-explicit-any */

import type { Inferer } from './../infer/inferer';

/** General context type must object with string key. */
type Ctx = Record<string, any>;

/** Context with previous value inside, used in pipe function. */
export type ContextPipe<T extends Ctx, V> = T & { value: V };

/** Pipe context injector. Used under procedure or pipe wrap function for resolving end value. */
export type PipeInjector<T extends Ctx, V> = { context: T; value: V };

/**
 * garda procedure types.
 *
 * @template O Output types.
 * @template I Array type of input arguments.
 * @template T Context types.
 */
export type Procedure<O, I extends any[], T extends Ctx> = ((
  ...input: I
) => O) & {
  /**
   * garda pipe parse utility function to parse output into new procedure with new value.
   *
   * ---
   *
   * ```js
   *
   * // Usage
   * const getContextCount = context
   *  .procedure((ctx) => ctx.getCount())
   *  .parse((data) => typeof data === 'number' ? data : 0)
   *
   * // type count is number
   * const count = getContextCount()
   *
   * // Usage with zod
   * const getZodContextCount = context
   *   .procedure((ctx) => ctx.getCount())
   *   .parse(z.number().parse) // or z.number().safeParse
   *
   * // type value is number
   * const value = getZodContextCount()
   *
   * ```
   *
   *  */
  parse: <V>(parser: (data: any) => V) => Procedure<V, I, T>;

  /**
   * garda pipe utility function to parse output into new procedure with new value.
   * Equal to parse but for handling promises output.
   *
   * ---
   *
   * ```js
   *
   * // Usage
   * const getUser = context
   *  .procedure(async (ctx, userid) => await ctx.fetchUserData(userid))
   *  .parseAsync((data) => {
   *    if (!data) return undefined
   *    // some validation or transformation here
   *    return data
   *  })
   *
   * const user = await getUser("some-id")
   *
   * // Usage with zod
   * const getUserData = context
   *   .procedure(async (ctx, userid) => await ctx.fetchUserData(userid))
   *   .parse(userDataSchema.parse) // or userDataSchema.safeParse | parseAsync | spa
   *
   * const userdata = await getUserData("some-id")
   *
   * ```
   *
   *  */
  parseAsync: <V>(
    parser: (data: any) => V
  ) => Procedure<Promise<Awaited<V>>, I, T>;

  /**
   * garda pipe utility function for infering output and transform to new output type.
   *
   * ---
   *
   * ```ts
   *
   * // Usage
   * const getUserState = context
   *  .procedure((ctx, key: string) => ctx.getLocalStorageState(key))
   *  .infer(inferer => ({
   *    status: inferer.string('user.status.label'),
   *    hasWife: inferer.string('user.partner.gender' === 'female',
   *    hasChild: inferer.number('user.children.count') > 0
   *  }))
   *
   * const state = getUserState('some-key')
   *
   * ```
   */
  infer: <V>(inferFn: (inferer: Inferer) => V) => Procedure<V, I, T>;

  /**
   * garda pipe utility function for infering output and transform to new output type.
   * Equal to infer but for handling promises output.
   *
   * ---
   *
   * ```ts
   *
   * // Usage
   * const getUserName = context
   *  .procedure(async (ctx, id: string) => await ctx.fetchUserData(id))
   *  .inferAsync(inferer =>
   *    inferer.string('data.fullName') + ' ' + inferer.string('data.lastName')
   *  )
   *
   * const name = await getUserName('some-user-id')
   *
   * ```
   */
  inferAsync: <V>(
    inferFn: (inferer: Inferer) => V
  ) => Procedure<Promise<Awaited<V>>, I, T>;

  /**
   * garda utility function for transforming and processing output from procedure.
   *
   * ---
   *
   * ```js
   *
   * // Usage
   * const getCount = context
   *  .procedure((ctx) => ctx.count)
   *
   * const doubleCount = getCount.pipe(({ value }) => value * 2)
   * const isCountEven = getCount.pipe(({ value }) => value % 2)
   *
   * const count = getCount() // return 1
   * const doubled = doubleCount() // return 2
   * const isEven = isCountEven() // return true
   *
   * ```
   */
  pipe: <V>(
    pipeFn: (context: ContextPipe<T, O>, ...input: I) => V
  ) => Procedure<V, I, T>;

  /**
   * garda utility for injecting updated context.
   *
   * useful when context is globally stored and updated,
   * but procedure need to be isolated.
   *
   * ---
   *
   * ```js
   *
   * // Example
   *
   * // isolated get data procedure with fetch
   * const getData = context
   *  .procedure((ctx) => fetch('/some-api', {
   *    headers: { "Authorization": `Bearer ${ctx.token}` }
   *  }))
   *
   * // get context and set token from cookie
   * const ctx = context()
   * ctx.token = cookie.get('Access-Token')
   * getData.inject(ctx) // inject updated context
   * const data = await getData()
   *
   * // or using context update utility
   * context.update((ctx) => {
   *  ctx.token = cookie.get('Access-Token)
   *  return ctx
   * })
   *
   * getData.inject(context)
   *
   * // or even better, directly update procedure context
   * getData.inject(ctx => {
   *  ctx.token = cookie.get('Access-Token)
   *  return ctx
   * })
   * ```
   */
  inject: (
    context: T | Context<T> | ((context: T) => void)
  ) => Procedure<O, I, T>;

  /**
   * garda utility for reseting context that has been injected to current procedure.
   */
  resetContext: () => Procedure<O, I, T>;
};

export type ProcedureOptions = {
  onAfterCalled: (value: any) => void;
  onBeforeCalled: (input: any[]) => void;
};

export type Context<T extends Record<string, any> = any> = (() => T) & {
  /**
   * context function, alias for context()
   *
   * > ! notice. Different treatment on immidiate false/true.
   * >
   * > on immidiate option false (default), `context()` and `context.get()`
   * > will create a fresh context everytime. But when immidiate is set to true,
   * > context is already initiated. So it's not going to be isolated context.
   * */
  get: () => T;

  /**
   * context utility for set context value
   *
   * > ! notice. Different tretment on immidiate false/true.
   * >
   * > on immidiate option true, when set is called. context isolated feature will be
   * > terminated because context is set to object that already initiated. And every update
   * > on context and it's object that context set to will be immidiately applied.
   * >
   * > Isolated feature of the context can be set again by calling `context.reset()`,
   * > and context will be set to initial context value.
   */
  set: (context: T) => void;

  /**
   * context utility for updating context value
   *
   * > ! notice. Different tretment on immidiate false/true.
   * >
   * > on immidiate option false (default). Given update will be treated as a hook function
   * > that only will be called when procedure is called as needed because of the isolated treatment.
   * >
   * > But when isolated feature is terminated by setting context through `context.set()`,
   * > update will be applied immidiately.
   */
  update: (fn: (context: T) => void) => void;

  /**
   * reset context to initial context initiation function.
   * this will remove all previous update to the context.
   * on immidiate set to true, context that is reseted will not point to previous initiated context.
   *
   * ```js
   * // Example
   * const context = createContext(() => {
   *  let count = 0
   *  return { setCount: (num) => count = num, count: () => count }
   * }, { immidiate: true })
   *
   * const ctx = context.get()
   *
   * ctx.setCount(10)
   * context().count() // this will return 10 too.
   *
   * context.update(c => c.setCount(20))
   * ctx.count() // this will return 20
   *
   * // on reset context
   * context.reset()
   * context().count() // will return 0
   * ctx.count() // will return 20
   *
   * ```
   */
  reset: () => void;

  /** create procedure from context */
  procedure: <I extends any[], V>(
    fn: (context: T, ...input: I) => V
  ) => Procedure<V, I, T>;
};

export type ContextOptions = {
  /**
   * Option for immidiately run context.
   * useful when context should be shared between procedure.
   *
   * When false, context will be isolated from each procedure and run when only needed.
   *
   * Default to false.
   */
  immidiate: boolean;
} & ProcedureOptions;

export type InferContext<C extends Context> = C extends Context<infer T>
  ? T
  : never;
