import type { DataType, DataRecord } from '../types.js';
import type { parseOptions } from './types.js';

/**
 * Inferer: safely infering data for managing and transforming data
 */
export class Inferer {
	_data: DataType;

	constructor(data: DataType) {
		this._data = data;
	}

	/**
	 * create error object
	 *
	 * @private
	 * @param {...string[]} tags
	 * @return {Error}
	 * @memberof Inferer
	 */
	private _createError(...tags: string[]) {
		return new Error(['[Infer Error]', ...tags].join(' '));
	}

	/**
	 * fallback handler for each infer type function
	 *
	 * @private
	 * @template T
	 * @param {string} path
	 * @param {string} type
	 * @param {DataType} value
	 * @param {parseOptions<T>} options
	 * @return {T}
	 * @memberof Inferer
	 */
	private _fallbackHandler<T>(
		path: string,
		type: string,
		value: DataType,
		options: parseOptions<T>
	) {
		if (options.parseHandler) return options.parseHandler(value);
		if (options.default !== undefined) return options.default;
		throw this._createError(`[path: '${path}']`, `[type: ${type}]`, `[value: ${value}]`);
	}

	/* ----------------------------- Public Utility ----------------------------- */

	/**
	 * Get raw value based on object path
	 */
	get(path = '') {
		if (path === '.' || path === '') return this._data;
		if (this._data === null || this._data === undefined || typeof this._data !== 'object')
			return undefined;

		// based on you-dont-need-lodash repository
		// @see https://github.com/you-dont-need/You-Dont-Need-Lodash-Underscore?tab=readme-ov-file#_get
		const travel = (rgexp: RegExp) =>
			path
				.split(rgexp)
				.filter(Boolean)
				.reduce((val, key) => {
					if (val === null || val === undefined || typeof val !== 'object') return undefined;
					if (Object.hasOwn(val, key)) {
						return (val as DataRecord)[key];
					} else {
						return undefined;
					}
				}, this._data);

		return travel(/[,[\]]+?/) ?? travel(/[,[\].]+?/);
	}

	/**
	 * check given path if value exist or not.
	 *
	 * @param {string} path
	 * @param {(data: DataType) => boolean} [validator=(v) => v !== undefined] default to `(v) => v !== undefined`
	 * @return {boolean}
	 */
	has(path: string, validator: (data: DataType) => boolean = (v) => v !== undefined) {
		const value = this.get(path);
		return validator(value);
	}

	/**
	 * Infer object path in runtime either value is string or throw an error
	 *
	 * @param {string} [path='']
	 * @param {parseOptions<string>} [options={}]
	 * @return {string}
	 * @memberof Inferer
	 */
	string(path = '', options: parseOptions<string> = {}) {
		const value = this.get(path);
		if (typeof value === 'string') return value;
		return this._fallbackHandler(path, 'string', value, options);
	}

	/**
	 * Infer object path in runtime either value is number (int|float) or throw an error
	 *
	 * @param {string} [path='']
	 * @param {(parseOptions<number> & { float?: boolean })} [options={}]
	 * @return {number}
	 * @memberof Inferer
	 */
	number(path = '', options: parseOptions<number> & { float?: boolean } = {}) {
		const value = this.get(path);
		if (typeof value === 'number') return value;
		if (typeof value === 'string') {
			if (options.float === true) {
				const parsed = parseFloat(value);
				if (!isNaN(parsed)) return parsed;
			} else {
				const parsed = parseInt(value);
				if (!isNaN(parsed)) return parsed;
			}
		}
		return this._fallbackHandler(path, 'number', value, options);
	}

	/**
	 * Infer object path in runtime either value is boolean or throw an error
	 *
	 * @param {string} [path='']
	 * @param {parseOptions<boolean>} [options={}]
	 * @return {boolean}
	 * @memberof Inferer
	 */
	boolean(path = '', options: parseOptions<boolean> = {}) {
		const value = this.get(path);
		if (typeof value === 'boolean') return value;
		if (typeof value === 'number') return !!value;
		if (typeof value === 'undefined') return false;
		if (value === null) return false;

		return this._fallbackHandler(path, 'boolean', value, options);
	}

	/**
	 * Infer object path in runtime either value is string | number or throw an error
	 *
	 * @param {string} [path='']
	 * @param {parseOptions<string | number>} [options={}]
	 * @return {string | number}
	 * @memberof Inferer
	 */
	stringOrNumber(path = '', options: parseOptions<string | number> = {}) {
		const value = this.get(path);
		if (typeof value === 'string') return value;
		if (typeof value === 'number') return value;
		return this._fallbackHandler(path, 'string|number', value, options);
	}

	/** alias for stringOrNumber */
	snum(path = '', options: parseOptions<string | number> = {}) {
		return this.stringOrNumber(path, options);
	}

	/**
	 * Infer object path in runtime is array then return as InfererArray object.
	 *
	 * @param {boolean} [defaultEmpty=false] when true return will fallback to empty array. On false will throw error. default: false.
	 */
	array(path = '', defaultEmpty = false) {
		const value = this.get(path);
		if (value instanceof Array) return new InfererArray(value);
		if (defaultEmpty === true) return new InfererArray([]);
		return this._fallbackHandler<InfererArray>(path, 'array', value, {});
	}

	/**
	 * Infer object path when path has value it will parse or transform with given function or throw an error when value is null or undefined.
	 */
	parse<T = DataType>(path: string, fn: (data: unknown) => T, options: parseOptions<T> = {}) {
		const value = this.get(path);
		if (value !== null || value !== undefined) return fn(value);
		return this._fallbackHandler(path, 'parse', value, options);
	}

	/**
	 * Infer object path and fallback to given preset value and type if type not match
	 */
	maybe<T>(path: string, preset: T) {
		const wrap = <A extends unknown[], Fn extends (...args: A) => unknown>(fn: Fn, ...args: A) => {
			try {
				return fn(...args) as ReturnType<Fn>;
			} catch (error) {
				return preset;
			}
		};

		// eslint-disable-next-line @typescript-eslint/no-this-alias
		const inferer = this;
		return {
			/** check if given path is a string or return preset value */
			string(options: parseOptions<string> = {}) {
				return wrap(inferer.string, path, options);
			},

			/** check if given path is a number or return preset value */
			number(options: parseOptions<number> & { float?: boolean } = {}) {
				return wrap(inferer.number, path, options);
			},

			/** check if given path is a boolean or return preset value */
			boolean(options: parseOptions<boolean> = {}) {
				return wrap(inferer.boolean, path, options);
			},

			/** check if given path is a string, number or return preset value */
			stringOrNumber(options: parseOptions<string | number> = {}) {
				return wrap(inferer.stringOrNumber, path, options);
			},

			/** check if given path has value then parse using given function or return preset value */
			parse<V>(fn: (data: unknown) => V, options: parseOptions<V> = {}) {
				return wrap(inferer.parse<V>, path, fn, options);
			},

			/** check if given path is array then return InfererArray or return preset value */
			array() {
				const value = inferer.get(path);
				if (value instanceof Array) return new InfererArray(value);
				return preset;
			},
		};
	}

	/**
	 * alias or shortcut for maybe with preset value null
	 */
	nullable(path = '') {
		return this.maybe(path, null);
	}
}

/**
 * Extended array object with typesafe infering and transform capability
 *
 * @export
 * @class InfererArray
 * @extends {Array}
 */
export class InfererArray extends Array {
	constructor(collection: DataType[]) {
		super();
		this.push(...collection);
	}

	/**
	 * loop each item and parse or transform item using given function.
	 *
	 * @param {(row: Inferer, index: number, arr: InfererArray) => T} fn
	 * @return {T[]} new transformed array
	 */
	inferEach<T = DataType>(fn: (row: Inferer, index: number, arr: InfererArray) => T) {
		const collection: T[] = [];
		for (let i = 0; i < this.length; i++) {
			const row = this[i];
			collection.push(fn(new Inferer(row), i, this));
		}

		return collection;
	}

	/**
	 * extended Array.at that return inferer object.
	 *
	 * @param {number} index The zero-based index of the desired code unit. A negative index will count back from the last item.
	 */
	inferAt(index: number) {
		return new Inferer(this.at(index));
	}
}