import type { SchemaNode, PermitNode } from './node.js';
import type { OmitUtilityProps } from './../types';
import type { PermitBranch } from './branch/permit.js';
import type { SchemaBranch } from './branch/schema.js';

export type SchemaScope = {
	$scope: string;
	$description: string;
};

/** Tree object for defining access control schema */
// eslint-disable-next-line @typescript-eslint/no-empty-interface
export interface SchemaTree<T = string> extends Record<string, SchemaTree<T> | T> {}

/** Generated tree schema type */
export type GeneratedSchema<T extends SchemaTree = SchemaTree> = T extends SchemaTree
	? {
			[P in keyof T]: T[P] extends SchemaTree ? Schema<GeneratedSchema<T[P]>> : SchemaNode;
		}
	: never;

// eslint-disable-next-line @typescript-eslint/no-explicit-any
export type RevertSchema<T> = T extends Record<string, any>
	? {
			[K in OmitUtilityProps<keyof T>]: T[K] extends SchemaNode ? string : RevertSchema<T[K]>;
		}
	: never;

// eslint-disable-next-line @typescript-eslint/no-explicit-any
export type SchemaExport<T> = T extends Record<string, any>
	? {
			[K in OmitUtilityProps<keyof T>]: T[K] extends SchemaNode
				? { $scope: string; $description: string }
				: SchemaExport<T[K]>;
		}
	: never;

/** Access control schema tree */
export type Schema<T extends GeneratedSchema> = T & SchemaBranch<T>;

// eslint-disable-next-line @typescript-eslint/no-explicit-any
export type PermitNodes<T = Record<string, any>> = T extends Record<string, any>
	? {
			[K in OmitUtilityProps<keyof T>]: T[K] extends SchemaNode
				? PermitNode
				: T[K] extends { $scope: string; $description: string }
					? PermitNode
					: PermitNodes<T[K]> & PermitBranch<T[K]>;
		}
	: never;

// eslint-disable-next-line @typescript-eslint/no-explicit-any
export type PermitSchema<T extends Record<string, any>> = PermitNodes<T> & PermitBranch<T>;

/**
 * Infering literal string to scopes
 *
 * @template T
 */
export type InferLiteralString<T, TD extends string> = T extends infer K
	? K extends string
		? K extends ''
			? `${TD}`
			: `${K}`
		: `${TD}`
	: `${TD}`;

export type ScopePrefixer<T, S> = T extends infer K
	? K extends string
		? K extends ''
			? ''
			: `${K}${InferLiteralString<S, '.'>}`
		: never
	: never;

export type UnionScopes<
	T,
	Prefix extends string = '',
	Spacer extends string = '.',
> = T extends Record<string, unknown>
	? {
			[K in keyof T]: K extends string
				? T[K] extends Record<string, unknown>
					? UnionScopes<
							T[K],
							`${ScopePrefixer<Prefix, Spacer>}${K}`,
							InferLiteralString<Spacer, '.'>
						>
					: `${ScopePrefixer<Prefix, Spacer>}${K}`
				: never;
		}[keyof T]
	: Prefix;

export type TransformSchema<T, V> = T extends SchemaTree
	? {
			[P in keyof T]: T[P] extends SchemaTree ? TransformSchema<T[P], V> : V;
		}
	: never;

export type SchemaOptions = {
	/** add prefix to scope. default to '' (none) */
	prefix: string;

	/** string to use for spacer for scope. default to '.' */
	spacer: string;
};