/* eslint-disable @typescript-eslint/no-inferrable-types */
import { describe, expect, test } from 'vitest';
import { createContext, createProcedure } from './../index';
import { z } from 'zod';

// import { createSchema } from '../../permit';
// import { z } from 'zod';

// const schema = {
//   secret: {
//     read: 'read value of secret',
//     write: 'update value of secret',
//   },
// };

// const access = createSchema(schema);

describe('Garda Unit Testing: General Utility', () => {
  const ctx = () => {
    let count = 0;
    return {
      add() {
        count++;
      },
      get count() {
        return count;
      },
      set count(value: number) {
        count = value;
      },
    };
  };

  /**
   * Test context that closed between call.
   * in every call, count should reset to 0.
   */
  test('closed context test', () => {
    const context = createContext(ctx);
    const getCount = context.procedure((ctx) => ctx.count);
    const counter = context.procedure((ctx) => {
      ctx.add();
      return ctx.count;
    });
    const doubleCounter = context.procedure((ctx) => {
      ctx.count = ctx.count + 2;
      return ctx.count;
    });

    const customCounter = context.procedure((ctx, num: number) => {
      ctx.count = ctx.count + num;
      return ctx.count;
    });

    // test get count must be still 0
    expect(getCount()).toBe(0);

    // test add count, but this should still be 0
    expect(counter()).toBe(1);
    expect(getCount()).toBe(0);

    expect(doubleCounter()).toBe(2);
    expect(getCount()).toBe(0);

    expect(customCounter(20)).toBe(20);
    expect(getCount()).toBe(0);
  });

  /**
   * Test context that shared between call.
   * count changed after call
   */
  test('shared context test', () => {
    const context = createContext(ctx, { immidiate: true });
    const getCount = context.procedure((ctx) => ctx.count);
    const counter = context.procedure((ctx) => {
      ctx.add();
      return ctx.count;
    });
    const doubleCounter = context.procedure((ctx) => {
      ctx.count = ctx.count + 2;
      return ctx.count;
    });

    const customCounter = context.procedure((ctx, num: number) => {
      ctx.count = ctx.count + num;
      return ctx.count;
    });

    expect(getCount()).toBe(0);

    expect(counter()).toBe(1);
    expect(getCount()).toBe(1);

    expect(doubleCounter()).toBe(3);
    expect(getCount()).toBe(3);

    expect(customCounter(20)).toBe(23);
    expect(getCount()).toBe(23);
  });

  /**
   * Pipe function with closed context should still carry
   * calculated value from previous pipe/procedure
   */
  test('pipe function utility test', () => {
    const context = createContext(ctx);
    const count = context.procedure((ctx) => ctx.count);
    const addCountBy2 = count.pipe((ctx) => ctx.value + 2);
    const subtractCount = (num: number) => {
      const calc = addCountBy2.pipe((ctx) => ctx.value - num);
      return calc();
    };

    expect(count()).toBe(0);
    expect(addCountBy2()).toBe(2);
    expect(subtractCount(1)).toBe(1);

    // when count called, count should still be 0
    expect(count()).toBe(0);
  });

  /** Infer chain function shoul work properly */
  test('infer function utility test', () => {
    const context = () => ({
      user: JSON.stringify({
        id: 'c8244705-a479-4af4-914c-e30574f2c7e6',
        firstName: 'Elmo',
        lastName: 'Rosenbaum',
      }),
    });

    const getUserName = createProcedure(context, (ctx) =>
      JSON.parse(ctx.user)
    ).infer((i) => `${i.string('firstName')} ${i.string('lastName')}`);

    expect(getUserName()).toBe('Elmo Rosenbaum');
  });

  /** Infer async chain function should work properly */
  test('infer async function utility test', async () => {
    const context = () => ({
      getUser: async () =>
        JSON.stringify({
          id: 'c8244705-a479-4af4-914c-e30574f2c7e6',
          firstName: 'Elmo',
          lastName: 'Rosenbaum',
        }),
    });

    const fetchUserName = createProcedure(context, async (ctx) =>
      JSON.parse(await ctx.getUser())
    ).inferAsync((i) => `${i.string('firstName')} ${i.string('lastName')}`);

    await expect(fetchUserName()).resolves.toBe('Elmo Rosenbaum');
  });

  /** Parser using zod chain function should work properly */
  test('parser function utility test', () => {
    const context = createContext(() => ({
      user: JSON.stringify({
        id: 'c8244705-a479-4af4-914c-e30574f2c7e6',
        firstName: 'Elmo',
        lastName: 'Rosenbaum',
      }),
    }));

    const userSchema = z.object({
      id: z.string(),
      firstName: z.string(),
      lastName: z.string(),
    });

    const zodParse = context
      .procedure((ctx) => JSON.parse(ctx.user))
      .parse(userSchema.parse);
    const zodSafeParse = context
      .procedure((ctx) => JSON.parse(ctx.user))
      .parse(userSchema.safeParse);

    expect(zodParse().firstName).toBe('Elmo');
    expect(zodSafeParse().success).toBeTruthy();
  });

  /** Parser async using zod chain function should work properly */
  test('parser async function utility test', async () => {
    const context = createContext(() => ({
      getUser: async () =>
        JSON.stringify({
          id: 'c8244705-a479-4af4-914c-e30574f2c7e6',
          firstName: 'Elmo',
          lastName: 'Rosenbaum',
        }),
    }));

    const userSchema = z.object({
      id: z.string(),
      firstName: z.string(),
      lastName: z.string(),
    });

    const zodParse = context
      .procedure(async (ctx) => JSON.parse(await ctx.getUser()))
      .parseAsync(userSchema.parse);
    const zodSafeParse = context
      .procedure(async (ctx) => JSON.parse(await ctx.getUser()))
      .parseAsync(userSchema.safeParse);
    const zodParseAsync = context
      .procedure(async (ctx) => JSON.parse(await ctx.getUser()))
      .parseAsync(userSchema.parseAsync);
    const zodSafeParseAsync = context
      .procedure(async (ctx) => JSON.parse(await ctx.getUser()))
      .parseAsync(userSchema.spa);

    const promiseParse = async () => (await zodParseAsync()).firstName;
    const promiseSafeParse = async () => (await zodSafeParseAsync()).success;

    await expect(promiseParse()).resolves.toBe('Elmo');
    await expect(promiseSafeParse()).resolves.toBeTruthy();
    await expect(zodParse()).resolves.toHaveProperty('firstName');
    await expect(zodSafeParse()).resolves.toHaveProperty('success');
  });

  /** Error throw testing  */
  test('on error handling', () => {
    const context = createContext(ctx);
    const onThrowError = context.procedure(() => {
      throw new Error('Error Throwed');
    });

    expect(onThrowError).toThrowError('Error Throwed');

    const deepPipeThrow = context
      .procedure(() => 1)
      .pipe(() => 2)
      .pipe(() => 3)
      .pipe(() => 4)
      .pipe(() => 5)
      .pipe(() => 6)
      .pipe(() => 7)
      .pipe(() => {
        throw new Error('8');
      });

    expect(deepPipeThrow).toThrowError('8');
  });

  /** Sequence work as expected */
  test('call sequence on pipe', () => {
    const runner = createProcedure(
      () => ({}),
      () => ['procedure']
    )
      .pipe(({ value }) => {
        value.push('pipe 1');
        return value;
      })
      .pipe(({ value }) => {
        value.push('pipe 2');
        return value;
      })
      .pipe(({ value }) => {
        value.push('pipe 3');
        return value;
      })
      .pipe(({ value }) => {
        value.push('joiner');
        return value.join(', ');
      });

    expect(runner()).toBe('procedure, pipe 1, pipe 2, pipe 3, joiner');

    const pipes = createProcedure(
      () => ({}),
      () => ['function']
    )
      .pipe(({ value }) => {
        value.push('pipe 1');
        return value;
      })
      .pipe(({ value }) => {
        value.push('pipe 2');
        return value;
      })
      .pipe(({ value }) => {
        value.push('pipe 3');
        return value;
      })
      .pipe(({ value }) => {
        value.push('joiner');
        return value.join(', ');
      });

    expect(pipes()).toBe('function, pipe 1, pipe 2, pipe 3, joiner');
  });

  /** Hook Utility test */
  test('onCalled hook utility test', () => {
    const afterContext = createContext(() => ({}), {
      onAfterCalled(value) {
        throw new Error(`value: ${value}`);
      },
    });

    const afterHook = afterContext.procedure(() => 'procedure');
    expect(afterHook).toThrowError('value: procedure');
    expect(afterHook.pipe(() => 'pipe 1')).toThrowError('value: pipe 1');
    expect(afterHook.pipe(() => 1).pipe(() => 2)).toThrowError('value: 2');

    const beforeContext = createContext(() => ({}), {
      onBeforeCalled(input) {
        throw new Error(`input: ${(input ?? []).join(', ')}`);
      },
    });

    const beforeHook = beforeContext.procedure(
      (_, a: number, b: string) => a + b
    );
    expect(() => beforeHook(2, 'dua')).toThrowError('input: 2, dua');
    expect(() => beforeHook(1, 'satu')).toThrowError('input: 1, satu');
  });

  test('Context utility test', () => {
    const counter = () => {
      let count = 0;
      return {
        get count() {
          return count;
        },
        set count(num: number) {
          count = num;
        },
        add: (num = 1) => (count += num),
      };
    };

    // create isolated context
    const isolatedCounter = createContext(counter);

    // create counter context copy
    const iCounter = isolatedCounter.get();

    // on icounter added, isolatedCounter should not be updated
    iCounter.add();
    expect(iCounter.count).toBe(1);
    expect(isolatedCounter().count).toBe(0);

    // on isolatedCounter update, iCounter will not be updated
    // this is because context by default should be isolated
    isolatedCounter.update((ctx) => ctx.add(2));
    expect(iCounter.count).toBe(1);
    expect(isolatedCounter.get().count).toBe(2);

    // retest setting iCounter
    // isolated counter must not be updated
    iCounter.count = 3;
    expect(iCounter.count).toBe(3);
    expect(isolatedCounter().count).toBe(2);

    // setting isolated counter to be targeting iCounter.
    // this means that both isolatedCounter context and iCounter will be linked
    // and isolation treatment is negated
    isolatedCounter.set(iCounter);
    iCounter.add();
    expect(iCounter.count).toBe(4);
    expect(isolatedCounter().count).toBe(4);

    // on update, iCounter will receive update too
    isolatedCounter.update((ctx) => ctx.add(3));
    expect(iCounter.count).toBe(7);
    expect(isolatedCounter().count).toBe(7);

    // retry another iCounter update
    iCounter.count = 2;
    expect(iCounter.count).toBe(2);
    expect(isolatedCounter().count).toBe(2);

    // on context reset, context should be isolated and the link is removed
    isolatedCounter.reset();
    expect(iCounter.count).toBe(2);
    expect(isolatedCounter().count).toBe(0);

    // ----

    // create shared coounter that link to every instance of context
    const sharedCounter = createContext(counter, { immidiate: true });

    // create another instance (facade) of shared counter
    const sCounter = sharedCounter.get();

    // on sCounter add, context should be updated too
    sCounter.add();
    expect(sCounter.count).toBe(1);
    expect(sharedCounter().count).toBe(1);

    sharedCounter.update((ctx) => ctx.add(2));
    expect(sharedCounter().count).toBe(3);
    expect(sCounter.count).toBe(3);

    sharedCounter.set(sCounter);
    sharedCounter.update((ctx) => (ctx.count = 0));
    expect(sharedCounter.get().count).toBe(0);
    expect(sCounter.count).toBe(0);

    sCounter.add();
    expect(sCounter.count).toBe(1);
    expect(sharedCounter().count).toBe(1);

    sharedCounter.reset();
    // shared counter reseted to default isolation
    // and not linked to sCounter
    expect(sharedCounter().count).toBe(0);
    // sCounter not affecting sharedCounter anymore
    expect(sCounter.count).toBe(1);
  });
});

// TODO: Create isolation feature unit test
describe('Garda Unit Testing: Advance Isolation Unit Test', () => {
  const counter = () => {
    let count = 0;
    return {
      get count() {
        return count;
      },
      set count(num: number) {
        count = num;
      },
      add: (num = 1) => (count += num),
    };
  };

  // create isolated context
  const context = createContext(counter);

  // generate counter context instances
  const counterCtx = context();

  // setup procedure
  const getCountValue = context.procedure((ctx) => ctx.count);
  const incrementCount = context.procedure((ctx, num: number = 1) => {
    ctx.add(num);
    return ctx.count;
  });

  test('on counter instance change, it will not affected other', () => {
    // this will only change itself
    counterCtx.add();

    // this will not change anything
    const inc = incrementCount(2);

    expect(inc).toBe(2);
    expect(counterCtx.count).toBe(1);
    expect(context().count).toBe(0);
    expect(getCountValue()).toBe(0);
  });

  test('on context linked to counterCtx, it will change all procedure', () => {
    context.set(counterCtx);
    const inc = incrementCount(2);

    expect(inc).toBe(3);
    expect(counterCtx.count).toBe(3);
    expect(context().count).toBe(3);
    expect(getCountValue()).toBe(3);

    // reset changes
    context.reset();
    expect(counterCtx.count).toBe(3);
    expect(context().count).toBe(0);
    expect(getCountValue()).toBe(0);
  });

  test('on procedure injected counterCtx, it wont change root context or any other procedure context.', () => {
    incrementCount.inject(counterCtx);
    const inc = incrementCount(4);

    expect(inc).toBe(7);
    expect(counterCtx.count).toBe(7);
    expect(context().count).toBe(0);
    expect(getCountValue()).toBe(0);

    // to reset injected procedure, inject it using original context
    incrementCount.inject(context);
    const inc2 = incrementCount(1);

    expect(inc2).toBe(1);
    expect(counterCtx.count).toBe(7);
    expect(context().count).toBe(0);
    expect(getCountValue()).toBe(0);
  });

  test('procedure inject that directly manipulate context.', () => {
    incrementCount.inject((ctx) => ctx.add(7));
    const inc = incrementCount(3);

    expect(inc).toBe(10);
    expect(counterCtx.count).toBe(7);
    expect(context().count).toBe(0);
    expect(getCountValue()).toBe(0);

    // reset context using provided utility
    incrementCount.resetContext();

    expect(incrementCount(1)).toBe(1);
    expect(counterCtx.count).toBe(7);
    expect(context().count).toBe(0);
    expect(getCountValue()).toBe(0);
  });
});
