import { describe, expect, test } from 'vitest';
import { infer } from './../index';
import { ZodError, z } from 'zod';

describe('Infer data test', () => {
	const data = JSON.parse(
		JSON.stringify({
			records: {
				count: 3,
				items: [
					{ id: 'one', value: 1, toggle: true },
					{ id: 'two', value: 2, toggle: false },
					{ id: 'three', value: 3, toggle: true },
				],
				last: 'three',
			},
			bool: false,
			list: ['one', 'two', 'three'],
			float: '10.000000777',
			map: [{ name: 'john' }, { name: 'smith' }, { name: 'burk' }, { name: 'leila' }],
		})
	);

	test('string infer success', () => {
		const v = infer(data);
		expect(v.string('records.last')).toBe('three');
	});

	test('string infer fail and throw error', () => {
		const v = infer(data);
		expect(() => v.string('records.count')).toThrowError('Infer Error');
		expect(() => v.string('records.count')).toThrowError('records.count');
		expect(() => v.string('records.count')).toThrowError('string');
	});

	test('number infer success', () => {
		const v = infer(data);
		expect(v.number('records.count')).toBe(3);
	});

	test('number infer fail and throw error', () => {
		const v = infer(data);
		expect(() => v.number('records.last')).toThrowError('Infer Error');
		expect(() => v.number('records.last')).toThrowError('records.last');
		expect(() => v.number('records.last')).toThrowError('number');
	});

	test('boolean infer success', () => {
		const v = infer(data);
		expect(v.boolean('bool')).toBe(false);
	});

	test('boolean infer fail and throw error', () => {
		const v = infer(data);
		expect(() => v.boolean('records')).toThrowError('Infer Error');
		expect(() => v.boolean('records')).toThrowError('records');
		expect(() => v.boolean('records')).toThrowError('boolean');
	});

	test('string or number infer success', () => {
		const v = infer(data);
		expect(v.stringOrNumber('float')).toBe('10.000000777');
	});

	test('string or number infer fail and throw error', () => {
		const v = infer(data);
		expect(() => v.stringOrNumber('list')).toThrowError('Infer Error');
		expect(() => v.stringOrNumber('list')).toThrowError('list');
		expect(() => v.stringOrNumber('list')).toThrowError('string|number');
	});

	test('array infer success', () => {
		const v = infer(data);
		const itemSchema = z.object({ id: z.string(), value: z.number(), toggle: z.boolean() });
		const list = itemSchema.array().parse(v.array('records.items'));

		expect(list).toBeInstanceOf(Array);
		expect(list[0].id).toBe('one');
	});

	test('array infer fail and throw error', () => {
		const v = infer(data);
		expect(() => z.string().array().parse(v.array('records'))).toThrowError('Infer Error');
		expect(() => z.string().array().parse(v.array('records'))).toThrowError('records');
		expect(() => z.string().array().parse(v.array('records'))).toThrowError('array');
	});

	test('array infer fail and throw zod error', () => {
		const v = infer(data);
		expect(() => z.string().array().parse(v.array('records.items'))).toThrowError(ZodError);
	});

	test('parse infer success', () => {
		const v = infer(data);
		const record = v.parse('records', z.object({ count: z.number(), last: z.string() }).parse);
		expect(record.count).toBe(3);
		expect(record.last).toBe('three');
	});

	test('parse infer using get in zod parse', () => {
		const v = infer(data);
		const record = z.object({ count: z.number(), last: z.string() }).parse(v.get('records'));
		expect(record.count).toBe(3);
		expect(record.last).toBe('three');
	});

	test('parse infer fail and throw error', () => {
		const v = infer(data);
		const fn = () => v.parse('map', z.object({ count: z.number(), last: z.string() }).parse);
		expect(fn).toThrowError(ZodError);
	});

	test('infer nullable', () => {
		const v = infer({});
		expect(v.nullable('a').string()).toBeNull();
		expect(v.nullable('a').number()).toBeNull();
		expect(v.nullable('a').boolean()).toBeNull();
		expect(v.nullable('a').stringOrNumber()).toBeNull();
		expect(v.nullable('a').array()).toBeNull();
		expect(v.nullable('a').parse(z.string().parse)).toBeNull();
	});

	test('infer through an array', () => {
		const v = infer(data);
		expect(v.string('records.items[0].id')).toBe('one');
		expect(v.string('map[0].name')).toBe('john');
		expect(v.string('list[0]')).toBe('one');

		expect(() => v.string('list[0].id')).toThrowError();
	});
});