import { assertType, describe, expect, test } from 'vitest';
import schemaObjectDeep from './schema-deep';
import schemaObject from './schema-shallow';
import {
  SchemaNode,
  SchemaBranch,
  createPermit,
  createSchema,
  type SchemaTree,
} from './../index';

describe('Schema unit testing', () => {
  const schema = createSchema(schemaObject);
  const schemaDeep = createSchema(schemaObjectDeep);

  test('utility typechecking', () => {
    expect(schema).toBeInstanceOf(SchemaBranch);
    expect(schema.extra).toBeInstanceOf(SchemaBranch);
    expect(schema.extra.sayHi).toBeInstanceOf(SchemaNode);
    expect(schema.extra.sayHi.$scope).toBeTypeOf('string');
    expect(schema.todo.action.$meta.scope).toBeTypeOf('string');
    assertType<string>(schema.todo.action.$model().create);
    assertType<SchemaTree>(schema.extra.$model());
  });

  test('deep typecheking', () => {
    assertType<string>(
      schemaDeep.long.action.action.action.action.action.action.action.action.$model()
        .create
    );
    assertType<SchemaTree>(
      schemaDeep.long.action.action.action.action.action.action.action.$model()
    );
  });

  test('able to define scope', () => {
    expect(schema.extra.$scope).toBe('extra');
    expect(schema.extra.sayHi.$scope).toBe('extra.sayHi');
    expect(
      schemaDeep.long.action.action.action.action.action.action.action.action
        .create.$scope
    ).toBe(
      'long.action.action.action.action.action.action.action.action.create'
    );

    const schemaWithPrefix = createSchema(schemaObject, {
      prefix: 'com.domain',
    });
    const customSchema = createSchema(schemaObject, {
      prefix: 'com.domain',
      spacer: ':',
    });

    expect(schemaWithPrefix.todo.action.create.$scope).toBe(
      'com.domain.todo.action.create'
    );
    expect(customSchema.todo.action.create.$scope).toBe(
      'com.domain:todo:action:create'
    );
  });

  test('Proper format on json stringify', () => {
    const smallSchema = createSchema({
      branch: { node: 'some action scope' },
    });

    const jsonSchema = JSON.stringify(smallSchema, undefined, 0);
    expect(jsonSchema).toBe(`{"branch":{"node":"some action scope"}}`);
    expect(JSON.stringify(schema.extra, undefined, 0)).toBe(
      `{"sayHi":"Enable user to say hi"}`
    );
  });

  test('failing schema format', () => {
    const schemaErrFormat = JSON.stringify({
      branch: {
        fail: 0,
        array: ['hoho'],
      },
    });
    const errSchema = () => createSchema(JSON.parse(schemaErrFormat));
    expect(errSchema).toThrowError();
  });
});

describe('Schema branch unit test', () => {
  const schema = createSchema(schemaObjectDeep);

  test('loop each schema node', () => {
    const found = [];
    schema.todo.$each((node) => found.push(node.$scope));
    expect(found.length).toBe(5);
  });

  test('export schema', () => {
    const test1 = schema.todo.$export();
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    expect((test1.schema.action as any).toJSON).toBeUndefined();
    expect(test1.schema.view.$description).toBeDefined();
    expect(test1.schema.view.$scope).toBe('todo.view');

    const test2 = schema.extra.$export();
    expect(JSON.stringify(test2.schema)).toBe(
      `{"sayHi":{"$scope":"extra.sayHi","$description":"Enable user to say hi"}}`
    );
  });

  test('refactor schema', () => {
    const s = createSchema(schemaObject);
    s.$refactor('com.domain', ':');
    expect(s.extra.sayHi.$scope).toBe('com.domain:extra:sayHi');
  });
});

describe('Permit Simple Use-Case Test', () => {
  const schema = createSchema(schemaObject);
  const permit = schema.$permit([
    'todo.view',
    'todo.action.toggle',
    'todo.action.edit',
    'todo.action.create',
    'extra.sayHi',
    'long.action.action.action.action.action.action.action.action.create',
  ]);

  test('has total of 5 access granted', () => {
    expect(permit.granted.length).toBe(5);
  });

  test('has been granted access to todo.action.create', () => {
    expect(permit.on.todo.action.create.$able).toBeTruthy();
  });

  test('denied access to delete todo', () => {
    expect(permit.on.todo.action.delete.$able).toBeFalsy();
  });

  test('able to change granted access of todo.action.edit to be denied', () => {
    permit.on.todo.action.edit.$toggle();
    expect(permit.on.todo.action.edit.$able).toBeFalsy();
  });

  test('set todo.view to be denied', () => {
    permit.on.todo.view.$able = false;
    expect(permit.on.todo.view.$able).toBeFalsy();
  });

  test('access that is not granted should be 3', () => {
    expect(permit.granted.length).toBe(3);
  });
});

describe('Permit Deep Use-case Test', () => {
  const schema = createSchema(schemaObjectDeep);
  const permit = schema.$permit([
    'long.action.action.action.action.action.action.action.action.create',
  ]);

  test('expect long child to success checking value', () => {
    expect(
      permit.on.long.action.action.action.action.action.action.action.action
        .create.$able
    ).toBeTruthy();
  });

  test('extract extra schema to be detached schema', () => {
    const extraPermit = schema.extra.$permit([]);
    expect(extraPermit.nodes.length).toBe(1);
    expect(extraPermit.on.sayHi.$scope).toBe('extra.sayHi');
    expect(extraPermit.on.sayHi.$able).toBe(false);
  });
});

describe('Permit Branch Utility Test', () => {
  const schema = createSchema(schemaObject);
  const permit = schema.$permit([
    'todo.view',
    'todo.action.toggle',
    'todo.action.edit',
    'todo.action.create',
  ]);

  test('expect "todo" branch strict value to be false', () => {
    expect(permit.on.todo.$able).toBe(false);
  });

  test('expect "todo" branch non-strict value to be true', () => {
    expect(permit.on.todo.$doable).toBe(true);
  });

  test('expect "extra" branch is disabled', () => {
    expect(permit.on.extra.$disable).toBe(true);
  });

  test('expect child of "extra" has all granted to be truth after changes', () => {
    permit.on.extra.sayHi.$toggle();
    expect(permit.on.extra.$able).toBe(true);
  });
});

describe('Permit Utility Test', () => {
  const schema = createSchema(schemaObject);
  const permit = schema.$permit(['todo.view']);

  test('loop utility test', () => {
    const found = [];
    permit.$each((node) => found.push(node.$scope));
    expect(found.length).toBe(6);
  });

  test('get child utility test', () => {
    expect(permit.child).toBeDefined();
    const child = permit.child('todo.action');
    expect(child.length).toBe(4);
    expect(
      child.map((n) => n.$scope).includes('todo.action.toggle')
    ).toBeTruthy();
  });

  test('reset utility test', () => {
    expect(permit.reset).toBeDefined();

    permit.on.todo.action.create.$able = true;
    expect(permit.on.todo.action.create.$able).toBeTruthy();

    permit.reset();
    expect(permit.on.todo.action.create.$able).toBeFalsy();
  });

  test('grant utility test', () => {
    expect(permit.grant).toBeDefined();

    permit.grant(
      'todo.action.create',
      'todo.action.edit',
      'todo.action.remove'
    );
    expect(permit.granted.length).toBe(3);
  });

  test('deny utility test', () => {
    expect(permit.deny).toBeDefined();

    permit.deny('todo.action.edit');
    expect(permit.granted.length).toBe(2);
  });

  test('granted utility test', () => {
    expect(permit.granted).toBeDefined();
    expect(permit.granted.length).toBe(2);
  });

  test('denied utility test', () => {
    expect(permit.denied).toBeDefined();
    expect(permit.denied.length).toBe(4);
  });

  test('permit tree utility test', () => {
    expect(permit.on).toBeDefined();
    expect(permit.on.extra.sayHi).toBeDefined();
    expect(permit.on.extra.sayHi.$scope).toBe('extra.sayHi');
    expect(permit.on.extra.sayHi.$description).toBe('Enable user to say hi');
    expect(permit.on.extra.sayHi.$able).toBeFalsy();
    expect(permit.on.extra.sayHi.$toggle).toBeDefined();

    permit.on.extra.sayHi.$toggle();
    expect(permit.on.extra.sayHi.$able).toBeTruthy();
  });

  test('Create permit from schema export', () => {
    const schema = createSchema(schemaObject);
    const data = schema.extra.$export();
    const json = JSON.stringify(data.schema);
    type SubSchema = typeof data.schema;

    const permit = createPermit<SubSchema>(JSON.parse(json), []);

    expect(permit.on.sayHi.$scope).toBe('extra.sayHi');
    expect(permit.on.sayHi.$able).toBeFalsy();

    permit.on.sayHi.$toggle();
    expect(permit.on.sayHi.$able).toBeTruthy();
  });
});
